#!/usr/bin/python
#
# Converter module for Master Of Orion 2 Graphics Viewer
# Copyright (c) Karol "Wolverine" Tomala 2007
# Contact: karol at karoltomala dot com
#
# This file is licensed under GPLv2
#

import Image,ImageDraw
from graphics import Graphic,Frame,Line
from types import TupleType

def savepng(infile,outfile,extpalfile=None):
    """savepng(infile,outfile,extpalfile=None): Converts graphic to indexed PNG. Takes infile as 
    graphic file, outfile as PNG to be saved and optionally external palette filename"""
    # Open graphic file
    gfx=Graphic(infile)
    # Get size
    wh=gfx.getwh()

    # Get internal palette
    pal=gfx.getipal()
    # If image has internal palette
    if (pal!=False):
	respal=[]
	# Create result palette for use with Python Imaging
	for i in range(0,len(pal)):
	    da,dr,dg,db=pal[i]
	    clrtuple=(int(dr*4),int(dg*4),int(db*4))
	    respal.extend(clrtuple)
    # If not, use external palette
    # FIXME: INTERNAL WITH EXTERNAL PALETTE MERGING!
    else:
	# Check if external palette file was specified
	if (extpalfile==None):
	    print "Error: No external palette files were specified!"
	    return False
	else:
	    # FIXME: EXTERNAL PALETTE LOADING HERE
	    print "FIXME: To be done"
	    return True
    
    # Go through each frame
    for fcount in range(0,gfx.countframes()):
	# Initialize new image
	im=Image.new('P',wh)
	# Set image palette
	im.putpalette(respal)
	# Get frame and it's contents
	frm=gfx.getframe(fcount)
	frmyindent=frm.getindent()

        # Initialize image drawing
	draw=ImageDraw.Draw(im)

	# Set relative x,y positions
        relx=0
	rely=frmyindent

	# Get through each pixel sequence (line) within frame
        for i in range(0,frm.countlines()):
	    # Get line
	    line=frm.getline(i)
	    # If it's normal line 
	    if (line.special==0):
		# Get line x indentiation
		relx+=line.indent
		# Parse each pixel
		for j in range(0,len(line.pixels)):
		    # Get pixel value
		    clr=line.pixels[j]
		    # Draw pixel with colour clr
		    draw.point((relx+j,rely),fill=clr)
		# Increment relative x position with number of pixels in the sequence
		relx+=len(line.pixels)
	    # If it's special line (indent change)
	    else:
		# Increment y indentiation and reset x indentiation
		rely+=line.indent
		relx=0

	# Save file for each frame
	savefile=outfile+'_frm_'+str(fcount)+'.png'
	im.save(savefile,'PNG',transparency=0)

def savegfx(infile,outfile,transparent=0,palcolors=0,palcolshift=0):
    """savegfx(infile,outfile,transparent=0,palcolors=0,palcolshift=0): Converts from PNG 
    to graphic file. Takes infile as PNG, outfile as graphic file, transparent color number,
    number of internal palette colors to use (0 means don't use internal palette),
    and number of palette colors shift for internal palette merging with external"""
    # Open file
    im=Image.open(infile)
    if (im.palette==None):
	print "Error: This PNG is not in 8 bit palette mode!"
	return
    # Get it's size
    w,h=im.size
    yindent=0

    # Create new graphic
    gfx=Graphic()
    # Set size
    gfx.setwh(w,h)

    # If we want to use internal palette
    if (palcolors>0):
	# Get internal palette
	lut=im.resize((256, 1))
	lut.putdata(range(256))
	lut=lut.convert("RGB").getdata()
	intpal=[]
	# Get only as much colors as needed for the image
	for i in range(0,palcolors):
	    r,g,b=lut[i]
	    # FIXME: Do something about DAC alpha in future
	    a=0
	    # Convert back to DAC RGB
	    da=int(a/4)
	    dr=int(r/4)
	    dg=int(g/4)
	    db=int(b/4)
	    # Append color tuple to internal palette
	    intpal.append((da,dr,dg,db))
	del lut
	# Set internal palette
	gfx.setipal(intpal,palcolshift)

    # Get frame
    frame=Frame()
    line=False
    lines=0
    # Get each line in the picture
    for y in range(0,h):
	xindent=0
	emptyscanline=True
	# Scan line
	for x in range(0,w):
	    # Get pixel
	    pix=im.getpixel((x,y))
	    # See if it's transparent
	    if (pix==transparent):
		# If we have line set
		if (line!=False):
		    # Add it to frame
		    frame.addline(line)
		    # Reset other data
		    xindent=0
		    line=False
		# Increase x indent
		xindent+=1
	    else:
		# If we had empty scanline to this point
		if (emptyscanline==True):
		    # If our frame didn't have scanlines
		    if (frame.countlines()==0):
			# Set frame y indent
			frame.setindent(yindent)
			yindent=0
		    else:
			# Create termination line
			line=Line()
			line.special=1
			line.indent=yindent
			yindent=0
			# And add it to our frame
			frame.addline(line)
			# Reset line
			line=False
		# If we don't have currently available line
		if (line==False):
		    # Create one
		    line=Line()
		    line.special=0
		    line.indent=xindent
		# Add pixel
		line.pixels.append(pix)
		# Set empty scan line flag to false
		emptyscanline=False
		
	# If we had line till the end of 
	if (line!=False):
	    # Add it to frame
	    frame.addline(line)
	    # Reset other data
	    xindent=0
	    line=False
	# Increase y indent
	yindent+=1
	
    # Add frame termination line
    line=Line()
    line.special=2
    line.indent=1000
    frame.addline(line)
    # Reset line
    line=False
    
    # Finally add frame to graphic
    gfx.addframe(frame)
    
    # Save file
    gfx.save(outfile)

#savepng(infile,'tester2')
#savegfx('picard2.png','officer.lbx_0023')
#savegfx('locutus2.png','officer.lbx_0036')
