#!/usr/bin/python
#
# Graphics module for Master Of Orion 2 Graphics Viewer
#
# Copyright (c) Karol "Wolverine" Tomala
# This file is licensed under GPLv2
#

import sys,os,string,types
from struct import unpack,pack

# ========= SUPPORT FUNCTIONS =========

def hex2bin(str):
    bin=['0000','0001','0010','0011',
        '0100','0101','0110','0111',
	'1000','1001','1010','1011',
        '1100','1101','1110','1111']
    bintxt=''
    for i in range(0,len(str)):
        bintxt+=bin[string.atoi(str[i],base=16)]
    return bintxt

# =========== LINE CLASS ===========

class Line:

    # ------------ CONSTRUCTOR ------------

    def __init__(self):
	"""Class constructor"""
	self.pixels=[]
	self.indent=0
	# Special is for new lines (changes indent from x to y)
	self.special=0

# =========== FRAME CLASS ===========

class Frame:

    # ------------ CONSTRUCTOR ------------

    def __init__(self):
	"""Class constructor"""	
	self.__lines=[]
	self.__indent=0

    # ------------ METHODS ------------

    def setindent(self,indent):
	"""setindent(indent): Sets y indentiation"""
	self.__indent=indent

    def getindent(self):
	"""getindent(): Returns y indentiation"""
	return self.__indent

    def addline(self,line):
	"""addline(line): Adds line"""
	self.__lines.append(line)

    def getline(self,num):
	"""getline(num): Returns line of number num"""
	if (num<0 or num>len(self.__lines)):
	    return False
	else:
	    return self.__lines[num]

    def countlines(self):
	"""countlines(): Returns number of lines"""
	return len(self.__lines)

    def create(self):
	"""create(): Creates raw frame data"""
	# Frame indicator
	frameind=pack('<H',1)
	# Frame indent
	yindent=pack('<H',self.getindent())
	lines=''
	# Create pixel sequences
	for i in range(0,self.countlines()):
	    # Get line
	    line=self.getline(i)
	    # Set pixel number in sequence
	    pixnum=len(line.pixels)
	    pixres=pack('<H',pixnum)
	    # Set sequence x indent
	    indent=pack('<H',line.indent)
	    # If we have actual pixel sequence (not special line terminator or end-of-data block)
	    if (pixnum>0):
		colors=''
		# Create each pixel
		for j in range(0,pixnum):
		    color=pack('<B',line.pixels[j])
		    colors+=color
		# If we have odd number of pixels round it to even by adding zero byte
		if (pixnum%2!=0):
		    colors+=pack('<B',0)
		# Create sequence data
		lineres=pixres+indent+colors
	    else:
		# Add sequence data
		lineres=pixres+indent
	    lines+=lineres
	result=frameind+yindent+lines
	# Check if frame divides by eight
	reslen=len(result)
	reslenmod=reslen%8
	if (reslenmod!=0):
	    zerocount=8-reslenmod
	    # Add additional zeroes
	    for i in range(0,zerocount):
		result+=pack('<B',0)
	# Return raw frame data
	return result

	
	
# =========== GRAPHIC CLASS ===========

class Graphic:

    # ------------ CONSTRUCTOR ------------

    def __init__(self,filename=None):
	"""Class constructor"""
	if (filename!=None):
	    self.__loaded=True
	    self.load(filename)
	else:
	    self.__loaded=False
	    self.__framelist=False
	    self.__frames=0
	    self.__delay=0
	    self.__flags={'junc':0,'ipal':0,'func':0,'fill':0,'nocomp':0}
    
    # ------------ INTERNAL FUNCTIONS ------------

    def __binflags(self,flags):
	"""binflags(flags): Converts raw flags to binary"""
	hflags=hex(flags)[2:]
	for i in range(len(hflags),4):
	    hflags='0'+hflags
	return hex2bin(hflags)

    def __rawflags(self):
	"""rawflags(): Converts flags dictionary to raw number"""
	binflags='00'+str(self.__flags['junc'])+str(self.__flags['ipal'])+str(self.__flags['func'])+str(self.__flags['fill'])+'0'+str(self.__flags['nocomp'])+'0'*8
	return int(string.atoi(binflags,2))
	
    def __parseflags(self,bflags):
	"""parseflags(bflags): Parses binary flags within the picture"""
	junction=int(bflags[2])
	ipalette=int(bflags[3])
	funccolor=int(bflags[4])
	fillbg=int(bflags[5])
	nocompress=int(bflags[7])
	return {'junc':junction,'ipal':ipalette,'func':funccolor,'fill':fillbg,'nocomp':nocompress}
	
    def __getheader(self,fp):
	"""getheader(fp): Reads file header"""
	# Get size
	self.__width,=unpack('<H',fp.read(2))
	self.__height,=unpack('<H',fp.read(2))

	# Skip two bytes
	fp.seek(6)
	
	# Get frames and their delay
	self.__frames,=unpack('<H',fp.read(2))
	self.__delay,=unpack('<H',fp.read(2))
	
	# Get binary flags
	flags,=unpack('<H',fp.read(2))
	bflags=self.__binflags(flags)
	self.__flags=self.__parseflags(bflags)
	
	# Frame offsets reading here
	byteformat='<'+'I'*(self.__frames+1)
	self.__offsets=unpack(byteformat,fp.read(4*(self.__frames+1)))

	# Get internal palette offset if available
	if (self.__flags['ipal']==1):
	    self.__flags['ipaloffset']=fp.tell()

    def __writeheader(self,fp,offsetar):
	"""writeheader(fp,offsetar): Writes header to file descriptor fp. Offsets array offsetar must be specified"""
	# Construct size
	size=pack('<H',self.__width)+pack('<H',self.__height)
	# Construct two null bytes
	nullstr=pack('<H',0)
	# Construct framecount and frame delay
	framedata=pack('<H',self.__frames)+pack('<H',self.__delay)
	# Flags construction
	flags=pack('<H',self.__rawflags())
	# Offsets construction
	offsets=''
	for i in range(0,len(offsetar)):
	    offsets+=pack('<I',offsetar[i])
	# Construct and write header
	header=size+nullstr+framedata+flags+offsets
	fp.write(header)

    def __createoffsets(self,rawdata,start):
	"""createoffsets(rawdata,start): Creates offsets for header inclusion from rawdata with starting adjustment start"""
	offsets=[]
	# First offset
	framelen=start
	offsets.append(framelen)
	# Other offsets
	for i in range(0,len(rawdata)):
	    framelen+=rawdata[i][0]
	    offsets.append(framelen)
	return offsets

    def __rawframes(self):
	"""rawframes(): Returns raw frame data"""
	framenum=self.countframes()
	result=[]
	for i in range(0,framenum):
	    frm=self.getframe(i)
	    rawdata=frm.create()
	    rawlen=len(rawdata)
	    result.append((rawlen,rawdata))
	return result

    def __getframes(self,fp):
	"""getframes(): Reads frames from file pointer fp"""
	self.__framelist=[]
	# Go through each offset
	for i in range(0,len(self.__offsets)-1):
	    # Position to offset start
	    fp.seek(self.__offsets[i])

	    # Read frame indicator
	    frameind,=unpack('<H',fp.read(2))
	    # If not proper frame indicator = try to go to next offset
	    if (frameind!=1):
		continue
	    
	    # Get y indent (relative y position) for frame
	    fyindent,=unpack('<H',fp.read(2))

	    # Create new frame
	    frm=Frame()
	    frm.setindent(fyindent)

	    # Begin reading pixel sequences for each frame
	    while (True):
		# Get pixel number
		pixels,=unpack('<H',fp.read(2))
	    
		# If we have 0 pixels, it's a special indication of either 
		# new pixel sequence or end of frame
		if (pixels==0):
		    # Get y indent (relative y position) of new line
		    yindent,=unpack('<H',fp.read(2))
		    
		    # If we've got yindent=1000 (E803)
		    if (yindent==1000):
			# We have encountered END OF DATA (frame) block
			linseq=Line()
			linseq.special=2
			linseq.indent=yindent
			frm.addline(linseq)
			# Go out of loop
			break
		    else:
			# We have normal y indent of the line
			# Create new line
			linseq=Line()
			linseq.special=1
			linseq.indent=yindent
			frm.addline(linseq)
			continue
		else:
		    # We've got number of pixels in the sequence
		    # Get x indent (relative x position) of the sequence
		    xindent,=unpack('<H',fp.read(2))
		    
		    # Read each 8bit pixel color value
		    byteformat='<'+'B'*pixels
		    colorarr=unpack(byteformat,fp.read(pixels))
		    
		    # Create new line sequence
		    linseq=Line()
		    linseq.special=0
		    linseq.indent=xindent
		    linseq.pixels=colorarr
		    frm.addline(linseq)
		    
		    # If pixel count is odd then round to even by reading additional byte
		    if (pixels%2!=0):
			# This byte should always be 0, but I'm not checking it
			fp.read(1)
	    # After getting frame data add it to frame array
	    self.addframe(frm)
	# If we've got no frames at all it's probably not a graphics file
	if (len(self.__framelist)==0):
	    self.__framelist=False

    def __getintpal(self,fp):
	"""getintpal(fp): Gets internal picture palette from file pointer fp"""
	# Position to the beginning of internal palette header
	fp.seek(self.__flags['ipaloffset'])
	# Color shift of the internal palette from the beginning of standard palette
	self.__colshift,=unpack('<H',fp.read(2))
	# Color number in internal palette
	colnum,=unpack('<H',fp.read(2))
	# Create temporary color array
	self.__ipaldacrgb=[]
	# Retrieve colors in DAC RGB
	for i in range(0,colnum):
	    dacrgb=unpack('<BBBB',fp.read(4))
	    self.__ipaldacrgb.append(dacrgb)

    def __createintpal(self):
	"""saveintpal(fp): Creates raw internal picture palette"""
	# Create palette color shift
	result=pack('<H',self.__colshift)
	# Create palette color number
	colnum=len(self.__ipaldacrgb)
	result+=pack('<H',colnum)
	# Create each color representation
	for i in range(0,colnum):
	    result+=pack('<BBBB',*self.__ipaldacrgb[i])
	return result

    def __writeframes(self,fp,framedata):
	"""writeframes(fp,framedata): Writes raw frames data framedata into file pointer fp"""
	for i in range(0,len(framedata)):
	    fp.write(framedata[i][1])

    # ------------ METHODS ------------

    def load(self,filename):
	"""load(filename): Loads graphic file"""
	self.__filename=filename
	fsize=os.path.getsize(filename)
	# Open the file
	fp=open(filename,'rb')

	# Get file header
	self.__getheader(fp)

	# Check for internal palette flag
	if (self.__flags['ipal']==1):
	    # Get internal palette
	    self.__getintpal(fp)
	if (self.__flags['func']==1):
	    sys.stderr.write('Notice: Functional color flag found. Image may not be accurately rendered.\n')
	    sys.stderr.flush()

	# Get frame data
	self.__getframes(fp)
	# Close the file
	fp.close()

    def save(self,filename):
	"""save(filename): Saves graphic file"""
	# Open the file
	fp=open(filename,'wb')
	# Create raw framedata array of tuples (length,data)
	framedata=self.__rawframes()
	
	# Create offsets
	start=12+(self.countframes()+1)*4
	# If there's internal palette
	if (self.__flags['ipal']==1):
	    # Adjust offset start by internal palette dimensions
	    start+=4+len(self.__ipaldacrgb)*4
	offsetar=self.__createoffsets(framedata,start)
	
	# Write header
	self.__writeheader(fp,offsetar)

	# If there's internal palette
	if (self.__flags['ipal']==1):
	    # Save internal palette
	    fp.write(self.__createintpal())

	# Write framedata
	self.__writeframes(fp,framedata)
	# Close the file
	fp.close()

    def getframe(self,num=0):
	"""getframe(num): Returns frame object of number num"""
	if (num>len(self.__framelist) or num<0):
	    return False
	else:
	    return self.__framelist[num]

    def addframe(self,frame):
	"""addframe(frame): Adds new frame"""
	if (type(frame)==types.InstanceType):
	    if (self.__framelist==False):
		self.__framelist=[]
	    if (self.__loaded==False):
		self.__frames+=1
	    self.__framelist.append(frame)

    def removeframe(self,num=0):
	"""removeframe(num): Removes frame object of number num"""
	if (num>len(self.__framelist) or num<0):
	    return False
	del(self.__framelist[num])
	
    def countframes(self):
	"""countframes(): Returns number of frames"""
	if (self.__framelist==False):
	    return False
	else:
	    return len(self.__framelist)

    def getwh(self):
	"""getwh(): Returns size tuple"""
	return (self.__width,self.__height)
    
    def setwh(self,w,h):
	"""setwh(w,h): Sets image size (only to be used with NEW IMAGES!)"""
	self.__width=w
	self.__height=h

    def getipal(self):
	"""getipal(): Returns internal palette"""
	if (self.__flags['ipal']==1):
	    return self.__ipaldacrgb
	else:
	    return False

    def setipal(self,ipaldacrgb,colshift=0):
	"""setipal(ipaldacrgb): Returns internal palette"""
	self.__flags['ipal']=1
	self.__colshift=colshift
	self.__ipaldacrgb=ipaldacrgb

    def getflags(self):
	"""getflags(): Returns flags"""
	return self.__flags

    def setflags(self,junction=0,ipalette=0,functional=0,fillback=0,nocompress=0):
	"""setflags(junction=0,ipalette=0,functional=0,fillback=0,nocompress=0): Sets image flags"""
	self.__flags={'junc':junction,'ipal':ipalette,'func':functional,'fill':fillback,'nocomp':nocompress}

    def getdelay(self):
	"""getdelay(): Returns delay"""
	return self.__delay

    def setdelay(self,delay):
	"""setdelay(delay): Sets delay between frames"""
	self.__delay=delay


# Break if called not as module
if __name__=='__main__':
    sys.stderr.write('Error: This is module. Run main.py instead!')
    sys.stderr.flush()
    sys.exit(1)
