#!/usr/bin/python
#
# Palette module for Master Of Orion 2 Graphics Viewer
#
# Copyright (c) Karol "Wolverine" Tomala
# This file is licensed under GPLv2
#

import sys,string
from struct import unpack
from types import ListType

# ========== PALETTE CLASS ==========

class Palette:
    
    # --------- CONSTRUCTOR -----------

    def __init__(self,filename,colnum=256):
	"""Class constructor"""
	self.__colnum=colnum
	self.__colors=[]
	# If we've got internal DAC RGB list (filename acts as color array)
	if (type(filename)==ListType):
	    for i in range(0,len(filename)):
		colorreg=filename[i]
		colorhex=self.__convrgb2hex(self.__convdacrgb(colorreg))
		self.__colors.append(colorhex)
	# If not just load the palette file
	else:
	    self.load(filename)
	
    # --------- INTERNAL FUNCTIONS -----------

    def __convdacrgb(self,dacrgb):
	"""condacrgb(dacreg): Converts dacrgb tuple to rgb tuple"""
	a,dr,dg,db=dacrgb
	dr*=4
	dg*=4
	db*=4
	return (dr,dg,db)
	
    def __convdacrgba(self,dacrgba):
	"""condacrgba(dacreg): Converts dacrgb tuple to rgba tuple"""
	da,dr,dg,db=dacrgb
	da*=4
	dr*=4
	dg*=4
	db*=4
	return (dr,dg,db,da)
	
    def __convrgb2hex(self,rgb):
	"""convrgb2hex(rgb): Converts rgb tuple to hex (24bit) representation"""
	r,g,b=rgb
	r=hex(int(r))[2:]
	g=hex(int(g))[2:]
	b=hex(int(b))[2:]
	if (len(r)==1):
	    r='0'+r
	if (len(g)==1):
	    g='0'+g
	if (len(b)==1):
	    b='0'+b
	return '#'+r+g+b

    def __convhex2rgb(self,hexstr):
	"""convhex2rgb(hexstr): Converts hexadecimal representation to RGB tuple"""
	hexstr=hexstr[1:]
	hr=string.atoi(hexstr[:2],16)
	hg=string.atoi(hexstr[2:4],16)
	hb=string.atoi(hexstr[4:],16)
	return (hr,hg,hb)
	

    # --------- METHODS ---------
	
    def load(self,filename):
	"""load(filename): Loads palette file"""
	self.__colors=[]
	fp=open(filename,'rb')
	for i in range(0,self.__colnum):
	    colorreg=unpack('<BBBB',fp.read(4))
	    colorhex=self.__convrgb2hex(self.__convdacrgb(colorreg))
	    self.__colors.append(colorhex)
	fp.close()

    def exportgimp(self,filename,name='Unknown'):
	"""exportgimp(filename): Exports palette as GIMP palette. GIMP palette name is taken 
	as optional argument"""
	result=[]
	result.append('GIMP Palette\n')
	result.append('Name: '+name+'\n')
	result.append('Columns: 0\n')
	result.append('#\n')
	for i in range(0,self.countcolors()):
	    clrtpl=self.__convhex2rgb(self.getcolor(i))
	    resline='%3d %3d %3d' % clrtpl
	    resline+='\tUntitled\n'
	    result.append(resline)
	fp=open(filename,'w')
	fp.writelines(result)
	fp.close()
    
    def countcolors(self):
	"""countcolors(): Returns number of colors within the palette"""
	return len(self.__colors)

    def getcolortuple(self,num)
	"""getcolortuple(num): Returns RGB color tuple of color number num"""
	if (num<0 or num>self.countcolors()):
	    return False
	else:
	    return self.__convhex2rgb(self.__colors[num])

    def gettuples(self):
	"""gettuples(): Get all colors as RGB tuple array"""
	result=[]
	for i in range(0,self.countcolors()):
	    result.append(self.getcolortuple(i))
	return result

    def getcolor(self,num):
	"""getcolor(num): Returns hexadecimal representation of the color which has 
	palette number num"""
	if (num<0 or num>self.countcolors()):
	    return False
	else:
	    return self.__colors[num]

    def getallcolors(self):
	"""getallcolors(): Returns whole palette (hexadecimal representation)"""
	return self.__colors

    def merge(self,pal,colshift):
	"""merge(pal): Merges this palette with other palette pal indented by color shift colshift"""
	# FIXME: Do this fucking palette merging -> needs internal structure representation rewrite


# Break if called not as module
if __name__=='__main__':
    sys.stderr.write('Error: This is module. Run main.py instead!')
    sys.stderr.flush()
    sys.exit(1)
